/*------------------------------------------------------------------------------------------------------------
 Version mit Motorblockadeerkennung  

 Realisiert mit Tiny25
 
 Interner Oszillator mit 8MHz
 
 Layout:
   RESET (ADC0)              1    8    +5V
   Motor-Spannung Neg (PB3)  2     7    PB2 (ADC1): Poti-Spannung
   Motor-Spannung Pos (PB4)  3     6    PB1: Blockierungsmeldung
   GND                       4     5    PB0: FET-Ausgang

 Input:
  PB4  ADC3 - Spannung an Pos. Motorseite (Versorgungsspannung)
  PB3  ADC2 - Spannung an neg. Motorseite (ber'm FET)
  PB2: ADC1 - Poti Spannung
 Output:
  PB1: Blockierungsmeldung 0=ok 1=blockiert 
  PB0  Ausgang zum FET
 
*/

#define F_CPU 8000000UL	/* CPU clock in Hertz */

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>

//------------------------------------------------------------------------------------------------------------
int messePotispannung(void){
  ADMUX=1; //ADC1/PB2 Poti-Spannung
  DIDR0 = (1<<ADC1D)|(1<<ADC2D); //Digitale Eingangsstufe abschalten
  ADCSRA = (1<<ADEN) |  //Enable ADC
           (1<<ADSC) |  //Start conversion
           (1<<ADIF) |   //Clear(!) interrupt flag
           (1<<ADPS2)|
           //(1<<ADPS1)|
           (1<<ADPS0);            //ADC Prescaler=32 - gibt 250KHz
           
  //Warte, bis Messung fertig
  while( !( ADCSRA & (1<<ADIF) ) )
  ;  
  ADMUX=6; //ADC2: PB4-PB3   (Motor-Spannung)
  return ADC;
}



//------------------------------------------------------------------------------------------------------------
int messeDiffMotor(void){
  //Motorspannung differentiell messen
  ADMUX=6; // Motor pos an PB4,  neg an PB3,  Gain 1x
  DIDR0 = (1<<ADC1D)|(1<<ADC2D); //Digitale Eingangsstufe abschalten
  ADCSRA = (1<<ADEN) |  //Enable ADC
           (1<<ADSC) |  //Start conversion
           (1<<ADIF) |   //Clear(!) interrupt flag
           (1<<ADPS2)|
           //(1<<ADPS1)|
           (1<<ADPS0);            //ADC Prescaler=32 - gibt 250KHz
           
  ADCSRB = 0;
  //Warte, bis Messung fertig
  while( !( ADCSRA & (1<<ADIF) ) )
  ;  
  ADMUX=1; //ADC1/PB2 Poti-Spannung
  return ADC;
}


//------------------------------------------------------------------------------------------------------------
void delay_25us(void){
static unsigned char i;
  i=10;
  do{
     asm volatile("NOP");
     asm volatile("NOP");
     asm volatile("NOP");
     asm volatile("NOP");

     asm volatile("NOP");
     asm volatile("NOP");
     asm volatile("NOP");
     asm volatile("NOP");
  }while (--i);  
}

//------------------------------------------------------------------------------------------------------------
int median(int value){
//ACHTUNG: Darf nur von 1 Stelle im Programm aufgerufen werden!
static int a,b,c, f1,f2,f3, x;
  f1=f2;
  f2=f3;
  f3=value;
  
  a=f1;
  b=f2;
  c=f3;

  if (a>b){
    x=a;
    a=b;
    b=x; 
  } //Jetzt ist a<=b

  if (b>c){
    x=c;
    c=b;
    b=x;
  }// b<=c
  if (a>b){
    x=a;
    a=b;
    b=x; 
  } //Jetzt ist a<=b

  return b;
}

//------------------------------------------------------------------------------------------------------------
int main(void){

  static int ist,soll,i, messungenU, messungenPoti;
  static int tastverhaeltnis;
  static int summeIst;
  static int summeSoll;
  static int delta;
  static int rp,ri;

  static int timeout, blockade;
 
  cli();
  
  DDRB = 3; //PB0 und 1 sind Ausgang
  PORTB = 0; //Motor aus; Alarm aus.
 
  tastverhaeltnis=0;
 
  rp=0;
  ri=0;
  messungenPoti=0;
  messungenU=0;
  summeIst=0;
  summeSoll=0;
  timeout=0;
  blockade=0;

  while(!blockade){  //Wird mit 300-400Hz durchlaufen
  
    summeSoll += messePotispannung();
    messungenPoti++;
    if (32==messungenPoti){
      soll = summeSoll>>5;
      summeSoll=0;
      messungenPoti=0;
    }
    
    messungenU++;
    if (messungenU > 5){
        ist = summeIst / messungenU;
        messungenU=0;
        delta = soll -32 - ist;

        summeIst=0;

        //P-Regler
        if (delta >  512){
          rp=100;
        } else
        if (delta < -512){
         rp=0;
        }else
        rp=50+delta/5;
    
        //I-Regler
        if ((delta >0) && (ri <  500)) ri += 1 + (delta>>5);
        if ((delta <0) && (ri > -500)) ri -= 1 - (delta>>5);
    }
    tastverhaeltnis = rp + ri/5;

    if (tastverhaeltnis >= 100){
      timeout++;
      if (timeout > 334){
        blockade=1;  //Nur durch Reset zu beheben!
      }
    }else{
      if (timeout>0) timeout--;
    }

    //Begrenzung
    if (tastverhaeltnis > 92) tastverhaeltnis = 92;
    if (tastverhaeltnis <  0) tastverhaeltnis =  0;

    //Messe Drehzahl
    summeIst += median(messeDiffMotor()); //Direkt vor dem Einschalten des Motors dessen Generatorspannung messen
    if (soll>31){  //Breiter Bereich fr AUS - damit auch sicher aus ist.
      PORTB |= 1; //Motor ein
      for (i=0; i< tastverhaeltnis; i++){  //EIN-Zeit
        delay_25us();  
      }  
    }  
    
    PORTB &= ~1; //Motor aus
    for (i=0; i< (100-tastverhaeltnis); i++){ //AUS-Zeit
      delay_25us();  
    }  
  }//While
  
  PORTB=2; //Motor aus, Alarm ein
  
  while(1) //Verharre bis zum Reset
  ;
  
  return 0;
}
// EOF----------------------------------------------------------------------------------------------------------
